"""Classes for grid search functionality."""

# TODO: implement Bayesian Optimization for hyperparameter tuning, then rename this dir to 'hyperparameter_tuning'

# Authors: Genevieve Hayes (modified by Andrew Rollings, Kyle Nakamura)
# License: BSD 3 clause

from .grid_search_mixin import GridSearchMixin
