"""Classes for defining fitness functions (i.e., optimization problems) for optimization algorithms."""

from .continuous_peaks import ContinuousPeaks

from .flip_flop import FlipFlop

from .four_peaks import FourPeaks

from .knapsack import Knapsack

from .max_k_color import MaxKColor

from .one_max import OneMax

from .queens import Queens

from .six_peaks import SixPeaks

from .travelling_salesperson import TravellingSalesperson

from .custom_fitness import CustomFitness
